/*
 * $Id: config.cpp,v 1.2 2002/09/01 22:50:34 fredrik Exp $
 * 
 * Keramik KWin client configuration module
 *
 * Copyright (C) 2002 Fredrik Hglund <fredrik@kde.org>
 *
 * Based on the Quartz configuration module,
 *     Copyright (c) 2001 Karol Szwed <gallium@kde.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the license, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
 
#include <kglobal.h>
#include <klocale.h>

#include <qcheckbox.h>
#include <qspinbox.h>
#include <qbuttongroup.h>
#include <qradiobutton.h>

#include "config.h"
#include "config.moc"

extern "C"
{
	QObject* allocate_config( KConfig* conf, QWidget* parent )
	{
		return ( new ActiveHeartConfig( conf, parent ) );
	}
}


/* NOTE: 
 * 'conf' 	is a pointer to the kwindecoration modules open kwin config,
 *			and is by default set to the "Style" group.
 *
 * 'parent'	is the parent of the QObject, which is a VBox inside the
 *			Configure tab in kwindecoration
 */

ActiveHeartConfig::ActiveHeartConfig( KConfig* conf, QWidget* parent )
	: QObject( parent )
{
	KGlobal::locale()->insertCatalogue("kwin_activeheart_config");
	c = new KConfig( "kwinactiveheartrc" );
	
	ui = new ActiveHeartConfigUI( parent );
	connect( ui->useShadowedText, SIGNAL(clicked()), SIGNAL(changed()) );
	connect( ui->roundBottomCorners, SIGNAL(clicked()), SIGNAL(changed()) );
	connect( ui->advancedColoringMode, SIGNAL(clicked()), SIGNAL(changed()) );
	connect( ui->useSeamlessBorder, SIGNAL(clicked()), SIGNAL(changed()) );
	connect( ui->titlebarHeightOffset, SIGNAL(valueChanged(int)),
		this, SIGNAL(changed()));
	load( conf );
	ui->show();
}


ActiveHeartConfig::~ActiveHeartConfig()
{
	delete ui;
	delete c;
}


// Loads the configurable options from the kwinrc config file
// It is passed the open config from kwindecoration to improve efficiency
void ActiveHeartConfig::load( KConfig* )
{
	c->setGroup("General");
	ui->useShadowedText->setChecked( c->readBoolEntry("UseShadowedText", true) );
	ui->roundBottomCorners->setChecked( c->readBoolEntry("RoundBottomCorners", false) );
	ui->advancedColoringMode->setChecked( c->readBoolEntry("AdvancedColoringMode", false) );
	ui->useSeamlessBorder->setChecked( c->readBoolEntry("UseSeamlessBorder", false) );
	ui->titlebarHeightOffset->setValue( c->readNumEntry("TitlebarHeightOffset", 0) );
}


// Saves the configurable options to the kwinrc config file
void ActiveHeartConfig::save( KConfig* )
{
	c->setGroup( "General" );
	c->writeEntry( "ShowAppIcons", true );
	c->writeEntry( "SmallCaptionBubbles", true );
	c->writeEntry( "LargeGrabBars", false);
	c->writeEntry( "UseShadowedText", ui->useShadowedText->isChecked() );
	c->writeEntry( "RoundBottomCorners", ui->roundBottomCorners->isChecked() );
	c->writeEntry( "AdvancedColoringMode", ui->advancedColoringMode->isChecked() );
	c->writeEntry( "UseSeamlessBorder", ui->useSeamlessBorder->isChecked() );
	c->writeEntry( "TitlebarHeightOffset", ui->titlebarHeightOffset->value() );
	c->sync();
}


// Sets UI widget defaults which must correspond to style defaults
void ActiveHeartConfig::defaults()
{
	ui->useShadowedText->setChecked( true );
	ui->roundBottomCorners->setChecked( false );
	ui->advancedColoringMode->setChecked( false );
	ui->useSeamlessBorder->setChecked( false );
	ui->titlebarHeightOffset->setValue(0);

	emit changed();
}

// vim: set noet ts=4 sw=4:
